<?php
/**
 * @plugin Listes électorales (prefix : Listes_electorales)
 * 
 * Licence GNU/GPL v3
 */

if (!defined('_ECRIRE_INC_VERSION')) return;

include_spip('inc/actions');
include_spip('inc/editer');

/*
 * list_calcul n'est pas un objet,
 * c'est un formulaire permettant de changer simultanément les données des listes et du scrutin
 * pour prendre en compte le dépouillement. Il s'agit d'une facilité d'écriture.
 * Son objectif est de permettre de modifier en même temps plusieurs objets.
*/

/*
 * Déclaration des champs du formulaire
 */
function formulaires_editer_list_calcul_saisies_dist($id_list_calcul='new', $retour='', $lier_trad=0, $config_fonc='', $row=array(), $hidden=''){
	$id_list_calcul = intval($id_list_calcul);
	$id_list_scrutin = intval(_request('id_list_scrutin'));

	// avoir les noms des listes et les voix
	$listes = array();
	$tous = sql_allfetsel('id_list_elec,titre,abrege,voix', 'spip_list_elecs', 'id_list_scrutin=' . $id_list_scrutin);
	// construire les saisies des voix qui se sont portées sur les listes
	foreach ($tous as $liste) {
		$listes[] = array(
			'saisie' => 'input',
			'options' => array(
				'nom' => 'id_list_elec-'.$liste['id_list_elec'],
				'label' => $liste['abrege'],
				'explication' => $liste['titre'],
				'defaut' => $liste['voix'],
				'obligatoire' => 'oui'
			),
			'verifier' => array(
				'type' => 'entier',
				'options' => array(
					'min' => 0,
				),
			),
		);
	}

	$scrutin = sql_fetsel('blancs,nuls,inscrits,siege,prime,seuil,repartition,acquisition,seuil_elegibilite,seuil_inclusion', 'spip_list_scrutins', 'id_list_scrutin=' . $id_list_scrutin);
	
	$saisies = array(
		array( // le fieldset 
			'saisie' => 'fieldset',
			'options' => array(
				'nom' => 'listes',
				'label' => _T('list_scrutin:fieldset_listes'),
				'icone' => 'identite-24.png',
			),
			'saisies' =>  $listes,
			),
		array( // le fieldset 
			'saisie' => 'fieldset',
			'options' => array(
				'nom' => 'participation',
				'label' => _T('list_scrutin:fieldset_participation'),
				'icone' => 'fiche-perso-24',
			),
			'saisies' => array( // les champs dans le fieldset 
				array( // les blancs
					'saisie' => 'input',
					'options' => array(
						'nom' => 'blancs',
						'label' => _T('list_scrutin:champ_blancs_label'),
						'explication' => _T('list_scrutin:champ_blancs_explication'),
						'obligatoire' => 'oui',
						'defaut' => $scrutin['blancs']
					),
					'verifier' => array(
						'type' => 'entier',
						'options' => array(
							'min' => 0,
						),
					),
				),
				array( // les nuls
					'saisie' => 'input',
					'options' => array(
						'nom' => 'nuls',
						'label' => _T('list_scrutin:champ_nuls_label'),
						'explication' => _T('list_scrutin:champ_nuls_explication'),
						'obligatoire' => 'oui',
						'defaut' => $scrutin['nuls']
					),
					'verifier' => array(
						'type' => 'entier',
						'options' => array(
							'min' => 0,
						),
					),
				),
			),
		),
		array( // le fieldset 
			'saisie' => 'fieldset',
			'options' => array(
				'nom' => 'information',
				'label' => _T('list_scrutin:fieldset_information'),
				'icone' => 'information-24',
			),
			'saisies' => array( // les champs dans le fieldset 
				array( // numéro unique du présent formulaire
					'saisie' => 'hidden',
					'options' => array(
						'nom' => 'id_list_calcul',
						'defaut' => $id_list_calcul,
					),
				),
					array( // numéro unique du scrutin
					'saisie' => 'hidden',
					'options' => array(
						'nom' => 'id_list_scrutin',
						'defaut' => $id_list_scrutin,
					),
				),
				array( // les inscrits
					'saisie' => 'input',
					'options' => array(
						'nom' => 'inscrits',
						'label' => _T('list_scrutin:champ_inscrits_label'),
						'explication' => _T('list_scrutin:champ_inscrits_explication'),
						'obligatoire' => 'oui',
						'defaut' => $scrutin['inscrits']
					),
					'verifier' => array(
						'type' => 'list_inscrits'
					),
				),
				array( // les sieges
					'saisie' => 'input',
					'options' => array(
						'nom' => 'siege',
						'label' => _T('list_scrutin:champ_siege_label'),
						'explication' => _T('list_scrutin:champ_siege_explication'),
						'obligatoire' => 'oui',
						'defaut' => $scrutin['siege']
					),
					'verifier' => array(
						'type' => 'entier',
						'options' => array(
							'min' => 1,
						),
					),
				),
				array( // la prime
					'saisie' => 'input',
					'options' => array(
						'nom' => 'prime',
						'label' => _T('list_scrutin:champ_prime_label'),
						'explication' => _T('list_scrutin:champ_prime_explication'),
						'obligatoire' => 'oui',
						'defaut' => $scrutin['prime']
					),
					'verifier' => array(
						'type' => 'list_prime'
					),
				),
				array( // le seuil
					'saisie' => 'input',
					'options' => array(
						'nom' => 'seuil',
						'label' => _T('list_scrutin:champ_seuil_label'),
						'explication' => _T('list_scrutin:champ_seuil_explication'),
						'obligatoire' => 'oui',
						'defaut' => $scrutin['seuil']
					),
					'verifier' => array(
						'type' => 'decimal',
						'options' => array(
							'min' =>'0.00',
							'max' => '99.99',
							'nb_decimales' => 2,
							'normaliser' => 'oui'
						)
					),
				),
				array( // la repartition
					'saisie' => 'radio',
					'options' => array(
						'nom' => 'repartition',
						'label' => _T('list_scrutin:champ_repartition_label'),
						'explication' => _T('list_scrutin:champ_repartition_explication'),
						'obligatoire' => 'oui',
						'datas' => array(
							'moyenne' => _T('list_scrutin:champ_repartition_moyenne_label'),
							'reste' => _T('list_scrutin:champ_repartition_reste_label'),
						),
						'defaut' => $scrutin['repartition']
					),
				),
				array( // l'acquisition
					'saisie' => 'radio',
					'options' => array(
						'nom' => 'acquisition',
						'label' => _T('list_scrutin:champ_acquisition_label'),
						'explication' => _T('list_scrutin:champ_acquisition_explication'),
						'obligatoire' => 'oui',
						'datas' => array(
							'on' => _T('list_scrutin:champ_acquisition_on_label'),
							'501exp' => _T('list_scrutin:champ_acquisition_501exp_label'),
						),
						'defaut' => $scrutin['acquisition']
					),
				),
				array( // le seuil_elegibilite
					'saisie' => 'input',
					'options' => array(
						'afficher_si' => '@acquisition@!="on"',
						'nom' => 'seuil_elegibilite',
						'label' => _T('list_scrutin:champ_seuil_elegibilite_label'),
						'explication' => _T('list_scrutin:champ_seuil_elegibilite_explication'),
						'obligatoire' => 'oui',
						'defaut' => $scrutin['seuil_elegibilite']
					),
					'verifier' => array(
						'type' => 'entier',
						'options' => array(
							'min' => 0,
						),
					),
				),
				array( // le seuil_inclusion
					'saisie' => 'input',
					'options' => array(
						'nom' => 'seuil_inclusion',
						'afficher_si' => '@acquisition@!="on"',
						'label' => _T('list_scrutin:champ_seuil_inclusion_label'),
						'explication' => _T('list_scrutin:champ_seuil_inclusion_explication'),
						'obligatoire' => 'oui',
						'defaut' => $scrutin['seuil_inclusion']
					),
					'verifier' => array(
						'type' => 'entier',
						'options' => array(
							'min' => 0,
						),
					),
				),
			),
		),
	);
	return $saisies;
}

/**
 * Identifier le formulaire en faisant abstraction des parametres qui ne representent pas l'objet édité
 */
function formulaires_editer_list_calcul_identifier_dist($id_list_calcul='new', $retour='', $lier_trad=0, $config_fonc='', $row=array(), $hidden=''){
	return serialize(array(intval($id_list_calcul)));
}

/**
 * Déclarer les champs postés et y integrer les valeurs par défaut
 */
function formulaires_editer_list_calcul_charger_dist($id_list_calcul='new', $retour='', $lier_trad=0, $config_fonc='', $row=array(), $hidden=''){

	# empêcher l'appel du formulaire sans id_list_scrutin
	if (!intval(_request('id_list_scrutin'))){
		return array('message_erreur', T_('list_scrutin:erreur_calcul_sans_id'));
	}

	# charger les valeurs du formulaire sans passer par formulaires_editer_objet_charger() qui cherchera une table spip_list_calculs inexistante
	$chargement = charger_fonction('editer_list_calcul_saisies','formulaires');
	$valeurs = $chargement();

	# sécurité sur l'autorisation détenue par l'auteur pour procéder au dépouillement et calcul
	include_spip('inc/autoriser');
	if (!autoriser('creer','list_calcul')){
		$valeurs['editable'] = false;
	}

	return $valeurs;
}

/**
 * Verifier les champs postés et signaler d'éventuelles erreurs
 */
function formulaires_editer_list_calcul_verifier_dist($id_list_calcul='new', $retour='', $lier_trad=0, $config_fonc='', $row=array(), $hidden=''){
	$erreurs = array();
	$erreurs = formulaires_editer_objet_verifier('list_calcul',$id_list_calcul);

	return $erreurs;
}

/**
 * Traiter les champs postés
 */
function formulaires_editer_list_calcul_traiter_dist($id_list_calcul='new', $retour='', $lier_trad=0, $config_fonc='', $row=array(), $hidden=''){
	$id_list_scrutin = _request('id_list_scrutin');
	$modifs = array();

	# enregistrement des valeurs modifiées du scrutin dans l'objet éditorial
	include_spip('action/editer_objet');
	$scrutin = array(
		'blancs' => _request('blancs'),
		'nuls' => _request('nuls'),
		'inscrits' => _request('inscrits'),
		'siege' => _request('siege'),
		'prime' => _request('prime'),
		'seuil' => _request('seuil'),
		'repartition' => _request('repartition'),
		'acquisition' => _request('acquisition'),
		'seuil_elegibilite' => _request('seuil_elegibilite'),
		'seuil_inclusion' => _request('seuil_inclusion'),
	);
	$erreur = objet_modifier('list_scrutin', $id_list_scrutin, $scrutin);
	if ($erreur) {
		return array ('message_erreur' => $erreur);
	} else {
		$modifs[] = array('id_list_scrutin' => $id_list_scrutin, 'data' => $scrutin);
	}
	
	# enregistrement des valeurs modifiées des listes présentées au scrutin dans les objets éditoriaux
	$tous = sql_allfetsel('id_list_elec,abrege,voix', 'spip_list_elecs', 'id_list_scrutin=' . $id_list_scrutin);
	foreach ($tous as $liste) {
		$voix = _request('id_list_elec-'.$liste['id_list_elec']);
		$erreur = objet_modifier('list_elec', $liste['id_list_elec'], array('voix' => $voix));
		if ($erreur) {
			return array ('message_erreur' => $erreur);
		} else {
			$modifs[] = array(
				'id_list_elec' => $liste['id_list_elec'],
				'data' => array('voix' => $voix)
			);
		}
	}
	
	# indiquer dans le log l'opération d'enregistrement des différents modifications
	spip_log(_T('list_scrutin:log_enregistrement_des_resultats', array(
			'id' => $id_list_scrutin,
			'id_auteur' => $GLOBALS['auteur_session']['id_auteur'],
			'nom' => $GLOBALS['auteur_session']['nom'],
			'modifs' => print_r($modifs,TRUE)
		)
	),
	'list_elec.' . _LOG_INFO_IMPORTANTE
	);

	# chargement et appel de la fonction de calcul des résultats
	$calcul_des_resultats = charger_fonction('calcul_des_resultats','formulaires');
	$repartition = $calcul_des_resultats($id_list_scrutin);

	# enregistrement des sièges obtenus dans la table auxiliaire
	foreach ($repartition as $id_list_elec => $tableau) {
		$champs = array(
			'id_list_scrutin' => $id_list_scrutin,
			'id_list_elec' => $id_list_elec,
			'pourcentage_des_suffrages' => $tableau['pourcentage_des_suffrages'],
			'repartition_au_quotient' => $tableau['repartition_au_quotient'],
			'prime' => $tableau['prime'],
			'attribution_moyenne' => $tableau['attribution_moyenne'],
			'class' => $tableau['class'],
			'specificites' => $tableau['specificites']
		);
		# modification
		if ($id_list_repartition = sql_getfetsel('id_list_repartition', "spip_list_repartitions", array ("id_list_elec=$id_list_elec", "id_list_scrutin=$id_list_scrutin"))){
			$champs = array_merge($champs, array ('id_list_repartition' => $id_list_repartition));
			if (!$retour = sql_updateq('spip_list_repartitions', $champs, 'id_list_repartition=' . intval($id_list_repartition))){
				spip_log(_T('list_scrutin:log_erreur_enregistrement_modification', array(
						'id_liste' => $id_list_elec,
						'id_scrutin' => $id_list_scrutin,
						'id_auteur' => $GLOBALS['auteur_session']['id_auteur'],
						'nom' => $GLOBALS['auteur_session']['nom'],
						'modifs' => print_r($champs,TRUE)
					)
				),
				'list_elec.' . _LOG_ERREUR
				);
				return array('message_erreur' => _T('list_scrutin:erreur_enregistrement_modification'));
			}
		} else {
		# insertion
			if (!$id = sql_insertq('spip_list_repartitions', $champs)){
				return array('message_erreur' => _T('list_scrutin:erreur_enregistrement_insertion', array('err' => $id)));
			}
		
		}
	}

	# rediriger vers la page de contenu et non pas d'édition (par défaut la page boucle sur elle-même.)
	return array(
		'message_ok' => 'Ok', // symbolique
		'redirect' => generer_url_ecrire('list_scrutin', 'id_list_scrutin=' . $id_list_scrutin),
	);
}

/**
 * Fonction de calcul des résultats électoraux
 *
 * @param  int    $id_list_scrutin
 *   Identifiant unique du scrutin dont on veut effectuer la répartition des sièges
 *   au regard des résultats
 * @return array $listes
 *   Description de la répartition des sièges entre les listes.
 *   Chaque liste est décrite par un tableau dont la clé est l'identifiant unique de la liste.
**/
function formulaires_calcul_des_resultats_dist($id_list_scrutin){

	if ($scrutin = sql_fetsel('siege,prime,seuil,repartition,acquisition,seuil_elegibilite,seuil_inclusion', 'spip_list_scrutins', 'id_list_scrutin=' . intval($id_list_scrutin))){

		$sieges_restants_a_pourvoir = $scrutin['siege'] - $scrutin['prime'];

		if ($toutes = sql_allfetsel('id_list_elec,voix', 'spip_list_elecs', 'id_list_scrutin=' . intval($id_list_scrutin))){

			# combien de listes a-t-on et quel est le total des suffrages exprimés ?

			$listes = array();
			$nb_listes = count($toutes);
			$suffrages_exprimes = 0;
			foreach ($toutes as $key => $value) {
				$suffrages_exprimes = $suffrages_exprimes + $value['voix'];
				$listes[$value['id_list_elec']] = array('voix' => $value['voix']);
			}

			# quelle est la liste arrivée en tête ?

			$resultats = array(); 
			$en_tête = array_keys($listes, max($listes)); 
			
			# quel est le pourcentage des suffrages pour chaque liste ?

			foreach ($toutes as $key => $value) {
				$listes[$value['id_list_elec']]['pourcentage_des_suffrages'] = round($value['voix'] / ($suffrages_exprimes/100),2);
				if ($scrutin['acquisition'] == '501exp'){
					if ($listes[$value['id_list_elec']]['pourcentage_des_suffrages'] > 50) {
						$listes[$value['id_list_elec']]['class'] = 'list_1';
						$acquis = 'oui';
					} elseif ($listes[$value['id_list_elec']]['pourcentage_des_suffrages'] >= intval($scrutin['seuil_elegibilite'])) {
						$listes[$value['id_list_elec']]['class'] = 'list_eligible';
					} elseif ($listes[$value['id_list_elec']]['pourcentage_des_suffrages'] >= intval($scrutin['seuil_inclusion'])) {
						$listes[$value['id_list_elec']]['class'] = 'list_inclue';
					}
				}
			}

			# si l'élection n'est pas acquise, on retourne le tableau sans répartition de siège

			if ($scrutin['acquisition'] == '501exp' and $acquis != 'oui') {
				return $listes;
			}


			# la prime majoritaire est donnée à la liste arrivée en tête

			if (count($en_tête)>1) {
					// il faut aller chercher la règle qui s'applique en cas d'égalité
			} else {
				$listes[$en_tête[0]]['prime'] = $scrutin['prime'];
			}
			
			# répartition des siège aux listes (ayant dépassé le seuil si spécifié)

			$retenues = array();
			foreach ($toutes as $key => $value) {
				$listes[$value['id_list_elec']]['pourcentage_des_suffrages'] = round($value['voix'] / ($suffrages_exprimes/100),2);
				if ($scrutin['seuil']) {
					if ($listes[$value['id_list_elec']]['pourcentage_des_suffrages'] >= $scrutin['seuil']) {
						$retenues[$value['id_list_elec']] = array('voix' => $value['voix']);
					}
				} else {
					$retenues[$value['id_list_elec']] = array('voix' => $value['voix']);
				}
				if ($scrutin['acquisition'] == '501exp') {
					if ($listes[$value['id_list_elec']]['pourcentage_des_suffrages'] > 50) {
						$retenues[$value['id_list_elec']] = array('acquis' => 'oui', 'class' => 'list_1');
					} elseif (isset($scrutin['seuil_elegibilite']) 
						//and intval($scrutin['seuil_elegibilite']) > 1
						and $listes[$value['id_list_elec']]['pourcentage_des_suffrages'] >= intval($scrutin['seuil_elegibilite'])) {
						$retenues[$value['id_list_elec']] = array('class' => 'list_eligible');
					} elseif (isset($scrutin['seuil_elegibilite']) 
						//and intval($scrutin['seuil_inclusion']) > 1
						and $listes[$value['id_list_elec']]['pourcentage_des_suffrages'] >= intval($scrutin['seuil_inclusion'])) {
						$retenues[$value['id_list_elec']] = array('class' => 'list_incluable');
					}
				}
			}


			# attribution des sièges aux listes ayant dépassées le seuil

			if ($sieges_restants_a_pourvoir) {
				# repartition au quotient électoral
				$quotient_electoral = round($suffrages_exprimes / $sieges_restants_a_pourvoir, 2);
				foreach ($retenues as $key => $value) {
					$listes[$key]['repartition_au_quotient'] =  floor($value['voix'] / $quotient_electoral);
					$sieges_restants_a_pourvoir = $sieges_restants_a_pourvoir - $listes[$key]['repartition_au_quotient'];
					$retenues[$key]['siege'] = $listes[$key]['repartition_au_quotient'];
				}

				if ($sieges_restants_a_pourvoir) {
					# attribution à la plus forte moyenne
					if ($scrutin['repartition'] == 'moyenne') {
						$sieges_pourvus = 0;
    					while ($sieges_pourvus < $sieges_restants_a_pourvoir){
							$plus_forte = plus_forte_moyenne($retenues);
							if (count($plus_forte)>1) {
								// il faut aller cherche qui a le plus de candidats
								$nb_candidats = array();
								foreach($plus_forte as $cle => $id_list_elec){
									$nb_candidats[$id_list_elec] = sql_countsel('spip_list_membres', array(
											"id_list_elec=$id_list_elec",
											"statut='publie'"
										)
									);
								}
								$plus_candidats = array_keys($nb_candidats,max($nb_candidats));
								// s'il y a toujours égalité, il y aura tirage au sort
								if (count($plus_candidats)>1) {
									foreach($plus_candidats as $cle => $id_list_elec){
										$ancien = unserialize($listes[$id_list_elec]['specificites']);
										if (!is_array($ancien)) {
											$ancien = array();
										}
										$ancien = array_merge($ancien, array('tirage_au_sort' => 1));
										$listes[$id_list_elec]['specificites'] = serialize($ancien);
									}
									$sieges_pourvus++;
									$retenues[$id_list_elec]['siege']++;
								}
								else {
									$listes[$plus_candidats[0]]['attribution_moyenne']++;
									$ancien = unserialize($listes[$id_list_elec]['specificites']);
									if (!is_array($ancien)) {
										$ancien = array();
									}
									$ancien = array_merge($ancien, array('candidats' => 1));
									$listes[$id_list_elec]['specificites'] = serialize($ancien);
									$sieges_pourvus++;
									$retenues[$plus_candidats[0]]['siege']++;
								}
							} else {
								$listes[$plus_forte[0]]['attribution_moyenne']++;
								$sieges_pourvus++;
								$retenues[$plus_forte[0]]['siege']++;
							}
						} // fin du while, boucle affectant les sièges restants
					}
				}
			}
		}
	}
	
	// construire les saisies des voix qui se sont portées sur les listes
	return $listes;
	//return array ('message_ok' => "<pre>".print_r($listes,true)."</pre>");
}

function plus_forte_moyenne($retenues){
    $moyenne = array();
	foreach($retenues as $key => $value){
		$moyenne[$key] = $value['voix'] / ($value['siege'] + 1) ;
	}
	$plus_forte = array_keys($moyenne,max($moyenne)); 
	spip_log(_T('list_scrutin:log_info_forte_moyenne', array(
				'moyenne' => print_r($moyenne,TRUE)
			)
		),
		'list_elec.' . _LOG_INFO_IMPORTANTE
	);
	return $plus_forte;
}